
#include <stdio.h>
#include <ncurses.h>
#include <unistd.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <sys/types.h>
#include <math.h>
#include <pthread.h>
#include <time.h>

#include "config.h"
#include "ncgui.h"
#include "pfregex.h"
#include "pfqueue.h"

int has_configpath;
int dig_suspend;
time_t dig_start;
int dig_limit;
int page_step;

void version() {
	printf ( "pfqueue version %s\n(C) 2004,2005 Stefano Rivoir\n", VERSION );
}

void usage() {
	version();
	printf ( "Usage: pfqueue [-hve] [-c postfix_config_dir] [-s autorefresh_seconds]"
		" [-m max_messages] [-p postfix_tools_dir]\n" );
}

void help() {
	WINDOW *w;
	char *ht;
	int c;

	w = newwin ( 25, 60, (LINES-25)/2,(COLS-60)/2 );

	nocbreak();
	cbreak();

	ht = help_text1;

	c = 0;
	while ( c!='q' ) {
		wclear ( w );
		mvwprintw ( w, 1, 0, ht );
		wnd_border ( w );
		wattron ( w, A_BOLD );
		mvwprintw ( w, 0, 28, "Help" );
		mvwprintw ( w, 0, 2, "pfqueue ver. " );
		mvwprintw ( w, 0, 15, VERSION );
		mvwprintw ( w, 23, 2, "Press q to exit, any key to next page" );
		wattroff ( w, A_BOLD );
		wrefresh ( w );
		c = wgetch ( w );
		ht = ( ht==help_text1 ? help_text2 : help_text1 );
	}
	delwin ( w );

	halfdelay ( delay_time*10 );
}

// Returns 1 if the file exists
int msg_fileex(int n) {
	return ( stat (queue[n].path, &foostat )!=-1);
}

// Reads one line from *fp
int freadl ( FILE *fp, char* buf, int buflen ) {
	int p = 0;

	if ( !fp )
		return 0;

	memset ( buf, 0, buflen );
	while ( p<buflen && fread ( buf+p, 1, 1, fp ) ) {
		if ( *(buf+p) == '\n' ) {
			*(buf+p) = '\0';
			return 1;
		}
		p++;
	}
	return 0;
}

void msg_cat() {
	WINDOW *w;
	FILE *p;
	char lines[MSG_MAX_ROWS][MSG_MAX_COLS];
	char b[BUF_SIZE];
	int j, i;
	int c;
	int fline;
	int pgstep;
	int msglines;
	
	if (!msg_fileex(CURMSG))
		return;

	if ( pf_version == PF_VERSION_1x ) {
		// postfix 1 does not support postcat -q
		if ( has_configpath )
			snprintf ( b, BUF_SIZE, "%s -c %s %s 2> /dev/null", 
				postcat_path, config_path, queue[CURMSG].path );
		else
			snprintf ( b, BUF_SIZE, "%s %s 2> /dev/null", 
				postcat_path, queue[CURMSG].path );
	} else {
		if ( has_configpath )
			snprintf ( b, BUF_SIZE, "%s -c %s -q %s 2> /dev/null", 
				postcat_path, config_path, queue[CURMSG].id );
		else
			snprintf ( b, BUF_SIZE, "%s -q %s 2> /dev/null", 
				postcat_path, queue[CURMSG].id );
	}
	
	p = popen ( b, "r" );
	if ( !p )
		return;
	j = 0;
	while ( j<MSG_MAX_ROWS && freadl(p, lines[j], MSG_MAX_COLS ))
		j++;
	msglines = j;
	pclose ( p );
	
	w = newwin ( LINES-4, COLS-4, 2, 2 );
	keypad ( w, TRUE );
	wnd_border (w);
	pgstep = LINES-4-2-2;
	
	wattron ( w, A_BOLD );
	mvwaddstr ( w, 0, 2, "Details for" );
	mvwaddstr ( w, 0, 14, queue[CURMSG].path );
	wattroff ( w, A_BOLD );

	c = 0;
	fline = 0;

	while ( c!='q' ) {
		j = fline;
		i = 2;
		while ( j<MSG_MAX_ROWS && i < LINES-4-1 ) {
			mvwaddnstr ( w, i, 2, lines[j++], COLS-4-3 );
			wclrtoeol ( w );
			mvwaddstr ( w, i++, COLS-4-1, "|" );
		}
		wrefresh ( w );

no_fill:
		c = wgetch (w);
		switch (c) {
		case 'g':
		case KEY_HOME:
			fline = 0;
			break;
		case 'G':
		case KEY_END:
			fline = msglines-pgstep;
			break;
		case KEY_PPAGE:
			if ( fline > pgstep )
				fline -= pgstep;
			else
				fline = 0;
			break;
		case KEY_NPAGE:
			if ( fline < (msglines-pgstep) )
				fline += pgstep;
			break;
		case KEY_UP:
			if ( fline )
				fline--;
			break;
		case KEY_DOWN:
			if ( fline<msglines-pgstep )
				fline++;
			break;
		case 'q':
			break;
		default:
			goto no_fill;
			break;
		}
	}
	delwin ( w );

	halfdelay ( delay_time*10 );
}

// Looks for a string at the beginning of the line
int flookfor ( FILE *fp, char* buf, int bufsize, char* w ) {
	bool b;
	int  wlen;
	char b2[BUF_SIZE];

	wlen = strlen ( w );
	b = FALSE;
	
	while ( !b && freadl ( fp, buf, bufsize ) )
		b = !strncmp ( buf, w, wlen );
	if ( b ) {
		memcpy ( b2, buf, bufsize );
		memcpy ( buf, b2+wlen, bufsize-wlen );
		*(buf+bufsize-1) = '\0';
	}
	return ( b );
}

// Checks if queue has the msg 'id'
int queue_hasid ( const char* id, int dd ) {
	if ( strncmp ( queue[dd].id, id, strlen(id) ) == 0 )
		return queue[dd].cached;
	else
		return 0;
}

// Looks for messages in the dir tree; recursive
int dir_dig( const char* basedir ) {
	DIR *bdir;
	struct dirent *dir;
	char full_path[BUF_SIZE];
	
	if ( NUMMSG_THREAD >= MSG_MAX )
		return -1;

	if ( dig_limit && ( time(NULL)-dig_start > dig_limit ) )
		return -1;
	
	bdir = opendir ( basedir );
	while ( bdir && NUMMSG_THREAD < MSG_MAX && ( dir = readdir ( bdir ) ) ) {

		if ( dig_limit && ( time(NULL)-dig_start > dig_limit ) )
			return -1;
	
		if ( dir->d_name[0] != '.' && ( dir->d_type == DT_DIR || dir->d_type == DT_LNK )) {
			snprintf ( full_path, sizeof(full_path), "%s/%s", basedir, dir->d_name );
			dir_dig ( full_path );
		} else if ( NUMMSG_THREAD < MSG_MAX && dir->d_type == DT_REG ) {
			strncpy ( queue_thread[NUMMSG_THREAD].id, dir->d_name, sizeof(queue_thread[NUMMSG_THREAD].id) );
			snprintf ( queue_thread[NUMMSG_THREAD].path, sizeof(queue_thread[NUMMSG_THREAD].path), "%s/%s",
				basedir, dir->d_name );
			queue_thread[NUMMSG_THREAD].cached = queue_hasid(dir->d_name, NUMMSG_THREAD);
			NUMMSG_THREAD++;
		} else if ( dir->d_name[0] != '.' && dir->d_type == DT_UNKNOWN ) {
			snprintf ( full_path, sizeof(full_path), "%s/%s", basedir, dir->d_name );
			stat ( full_path, &foostat );
			if ( S_ISDIR(foostat.st_mode) || S_ISLNK(foostat.st_mode)) {
				dir_dig ( full_path );
			} else if ( NUMMSG_THREAD < MSG_MAX && S_ISREG(foostat.st_mode) ) {
				strncpy ( queue_thread[NUMMSG_THREAD].id, dir->d_name, sizeof(queue_thread[NUMMSG_THREAD].id) );
				snprintf ( queue_thread[NUMMSG_THREAD].path, sizeof(queue_thread[NUMMSG_THREAD].path), "%s/%s",
					basedir, dir->d_name );
				queue_thread[NUMMSG_THREAD].cached = queue_hasid(dir->d_name, NUMMSG_THREAD);
				NUMMSG_THREAD++;
			}
		}
		
	}
	if ( bdir )
		closedir ( bdir );

	return 0;
}

// Do thread dirty work
void* dir_dig_start(void *arg) {
	char buf[BUF_SIZE];
	int i;
	
	while ( 1 ) {
		if ( !dig_suspend && DO_SCAN ) {
			if ( dig_limit )
				dig_start = time(NULL);
			NUMMSG_THREAD = 0;
			snprintf ( buf, sizeof(buf), "%s/%s", queue_path, q_names[CURQ] );
			dir_dig ( buf );
			for ( i=0; i<NUMMSG_THREAD; i++ ) {
				if ( !queue_thread[i].cached )
					memcpy ( queue+i, queue_thread+i, sizeof(struct msg_t) );
			}
			NUMMSG = NUMMSG_THREAD;
			dig_lastqueue = CURQ;
		}
		sleep ( delay_time );
	}
	pthread_mutex_unlock ( &dig_mutex );
	pthread_exit(NULL);
}

// Launches the dig thread
int fill_list() {
	if ( pthread_mutex_trylock(&dig_mutex)!=0 )
		return -1;
	pthread_create ( &dig_thread, NULL, dir_dig_start, NULL );
	return 0;
}

void goto_msg(int i) {
int j;
	PREVROW = CURROW;
	CURMSG  = i;
	CURROW  = (i-FIRSTMSG)+FIRSTROW;

	if ( CURROW<FIRSTROW ) {
		CURROW = FIRSTROW;
		FIRSTMSG = CURMSG;
	}
	if ( CURROW>MAXROW ) {
		j = CURROW-MAXROW;
		FIRSTMSG += j;
		CURROW = MAXROW;
	}
}

void goto_next() {
	if ( CURMSG<NUMMSG-1 )
		goto_msg( CURMSG+1 );
}

void goto_prev() {
	if ( CURMSG )
		goto_msg( CURMSG-1 );
}

void goto_first() {
	goto_msg(0);
}

void goto_last() {
	if ( NUMMSG )
		goto_msg( NUMMSG-1 );
}
void goto_nextpage() {
	if ( CURMSG+page_step < NUMMSG-1 )
		goto_msg ( CURMSG+page_step );
	else
		goto_msg ( NUMMSG-1 );
}
void goto_prevpage() {
	if ( CURMSG > page_step )
		goto_msg ( CURMSG-page_step );
	else
		goto_msg ( 0 );
}

// Extracts header data from the message
int msg_head ( int m ) {
	char buf[BUF_SIZE];
	char b[BUF_SIZE];
	int f1, f2, f3, l1, l2;
	char *s1, *s2;
	
	FILE *p;
	
	if ( !msg_fileex(m) ) {
		strcpy ( queue[m].id,   MSG_MOVED );
		strcpy ( queue[m].from, MSG_MOVED );
		strcpy ( queue[m].to,   MSG_MOVED );
		queue[m].cached = 0;
		return 1;
	}

	if ( pf_version == PF_VERSION_1x ) {
		// postfix 1 does not support postcat -q
		if ( has_configpath )
			snprintf ( b, BUF_SIZE, "%s -c %s %s 2> /dev/null", 
				postcat_path, config_path, queue[m].path );
		else
			snprintf ( b, BUF_SIZE, "%s %s 2> /dev/null", 
				postcat_path, queue[m].path );
	} else {
		if ( has_configpath )
			snprintf ( b, BUF_SIZE, "%s -c %s -q %s 2> /dev/null", 
				postcat_path, config_path, queue[m].id );
		else
			snprintf ( b, BUF_SIZE, "%s -q %s 2> /dev/null", 
				postcat_path, queue[m].id );
	}

	p = popen ( b, "r" );
	if ( !p ) {
		strcpy ( queue[m].from, "cant popen" );
		strcpy ( queue[m].to,   "cant popen" );
		return 2;
	}
	
	f1 = f2 = f3 = 0;
	
	if ( use_envelope ) {
		l1 = ENVELOPE_FROM_LEN;
		l2 = ENVELOPE_TO_LEN;
		s1 = ENVELOPE_FROM;
		s2 = ENVELOPE_TO;
	} else {
		l1 = HEADER_FROM_LEN;
		l2 = HEADER_TO_LEN;
		s1 = HEADER_FROM;
		s2 = HEADER_TO;
	}
	
	while ( (f1==0||f2==0||f3==0) && freadl ( p, buf, BUF_SIZE ) ) {
		if ( !f1 && ( !strncmp ( buf, s1, l1 ) ) ) {
			memcpy ( queue[m].from, buf+l1, sizeof(queue[m].from) );
			if ( !strlen(queue[m].from) )
				strcpy ( queue[m].from, ENVELOPE_NULL_SENDER );
			f1++;
		}
		if ( !f2 && ( !strncmp ( buf, s2, l2 ) ) ) {
			memcpy ( queue[m].to,   buf+l2, sizeof(queue[m].to) );
			f2++;
		}
		if ( !f3 && !strncmp ( buf, HEADER_SUBJECT, HEADER_SUBJECT_LEN ) ) {
			memcpy ( queue[m].subj, buf+HEADER_SUBJECT_LEN, sizeof(queue[m].subj) );
			f3++;
		}
	}
	pclose ( p );

	return 3;
}

// Fetch data (if not cached) for current message
int msg_fetch( int n ) {
	FILE *p;
	char buf[BUF_SIZE];
	char buf2[BUF_SIZE];
	char *c;
	int  res;
	
	if ( !msg_fileex ( n ) ) {
		strcpy ( queue[n].from, MSG_MOVED );
		strcpy ( queue[n].to  , MSG_MOVED );
		strcpy ( queue[n].subj, MSG_MOVED );
		queue[n].cached = 0;
		return 11;
	}
	if ( queue[n].cached )
		return 10;

	res = msg_head ( n );

	if ( CURQ == Q_DEFERRED ) {
		// replace 'deferred' with 'defer'
		c = strstr ( queue[n].path, "deferred" );
		if ( c ) {
			memset ( buf, 0, sizeof(buf) );
			strncpy ( buf, queue[n].path, c - queue[n].path );
			sprintf ( buf2, "%sdefer%s", buf, c+8 );
		}
		p = fopen ( buf2, "r" );
		if ( !p )
			strcpy ( queue[n].cause, "NONE" );
		else {
			if ( pf_version > PF_VERSION_20 )
				flookfor ( p, queue[n].cause, sizeof(queue[n].cause), "reason=" );
			else
				freadl ( p, queue[n].cause, sizeof(queue[n].cause) );
		}

		if ( p )
			fclose ( p );
	}
	if ( strlen(queue[n].to) )
		queue[n].cached = 1;
	return res;
}

// Performs an action on the message
void msg_action_do ( int i, int act ) {
	char b[BUF_SIZE];
	char o;
	
	if ( !ISROOT )
		return;
	if ( !NUMMSG )
		return;

	switch ( act ) {
	case MSG_DELETE:
		o = 'd';
		break;
	case MSG_HOLD:
		o = 'h';
		break;
	case MSG_RELEASE:
		o = 'H';
		break;
	case MSG_REQUEUE:
		o = 'r';
		break;
	default:
		return;
	}

	if ( has_configpath )
		snprintf ( b, BUF_SIZE, "%s -c %s -%c %s 2>/dev/null", 
			postsuper_path, config_path, o, queue[i].id );
	else
		snprintf ( b, BUF_SIZE, "%s -%c %s 2>/dev/null", 
			postsuper_path, o, queue[i].id );
	system ( b );
}

// Tags all messages
void msg_tagall() {
int i;
	for ( i=0; i<NUMMSG; i++ ) {
		msg_fetch ( i );
		queue[i].tagged = TRUE;
	}
	NUMTAG = NUMMSG;
}

// Resets tag flag on all messages
void msg_tagreset() {
int i;
	for ( i=0; i<NUMMSG; i++ )
		queue[i].tagged = FALSE;
	WRK_TAGGED = FALSE;
	NUMTAG = 0;
}

// Reset the cache flag of the messages
void msg_cachereset() {
int i;
	for ( i=0; i<NUMMSG; i++ )
		queue[i].cached = 0;
}

// Wrapper
void msg_action ( int act ) {
int i;

	if ( !ISROOT )
		return;

	if ( !WRK_TAGGED && !msg_fileex(CURMSG) )
		return;

	if ( ASK_CONFIRM && !wnd_confirm ( a_names[act] ) )
		return;

	if ( WRK_TAGGED ) {
		dig_suspend = 1;
		for ( i = 0; i<NUMMSG; i++ ) {
			if ( queue[i].tagged )
				msg_action_do ( i, act );
		}
		msg_tagreset();
		dig_suspend = 0;
		CURROW = FIRSTROW;
		CURMSG = 0;
	} else
		msg_action_do ( CURMSG, act );
	return;
}


// Sets the list window title
void win_listtitle() {
	wattron ( lwnd, A_BOLD );

	char buf[BUF_SIZE];
	sprintf ( buf, "Queue: '%s', %d message%s, %d tagged %s %s %s",
			q_names[CURQ], NUMMSG, 
			( NUMMSG != 1 ? "s" : "" ),
			NUMTAG,
			( WRK_TAGGED ? "*Commands work on tagged*" : "" ),
			( ASK_CONFIRM ? "" : "*Confirmation disabled*" ),
			( DO_SCAN ? "" : "*Scan suspended*" ) );
	mvwprintw ( lwnd, 0, 1, buf );

	wclrtoeol ( lwnd );
	wattron ( lwnd, A_BOLD );
	mvwaddstr ( lwnd, 1, 1, "ID" );
	mvwaddstr ( lwnd, 1, COL1, "From" );
	mvwaddstr ( lwnd, 1, COL2, "To" );
	wclrtoeol ( lwnd );
	wattroff ( lwnd, A_BOLD );
	wnoutrefresh ( lwnd );
}

// Sizes the window according to the (new) screen size
void win_resize () {

	int h;

	if ( CURQ == Q_DEFERRED )
		h = 8;
	else
		h = 6;
	
	MAXROW = LINES-h-1;

	if ( mwndb )
		delwin ( mwndb );
	if ( lwnd )
		delwin ( lwnd );
	if ( mwnd )
		delwin ( mwnd );
	if ( bline )
		free ( bline );

	mainwnd = initscr();

	bline = (char*)malloc(sizeof(char)*COLS);
	memset ( bline, ' ', sizeof(char)*COLS-1 );
	*(bline+COLS-1) = 0;

	mwndb = newwin ( h, COLS, LINES-h, 0 );
	lwnd  = newwin ( LINES-h, COLS, 0, 0 );
	mwnd  = newwin ( h-2, COLS-2, LINES-h+1, 1 );

	if ( use_envelope )
		wnd_clear ( mwndb, "Message (from envelope)" );
	else
		wnd_clear ( mwndb, "Message (from headers)" );

	COL1 = 14;
	COL2 = ( COLS - 16 ) / 2 - 1;
	
	wnoutrefresh ( mwndb );

	page_step = MAXROW -2;

	doupdate();
}

// Clears the cursor in the previous position and displays the current one
void show_cursor() {
	mvwaddstr ( lwnd, PREVROW, 0, " " );
	mvwaddstr ( lwnd, CURROW,  0, ">" );
}

// Changes the current queue
void queue_change( int q ) {
	CURROW = PREVROW = FIRSTROW;
	FIRSTMSG = 0;
	CURMSG = 0;
	CURQ = q;
	NUMTAG = 0;
	WRK_TAGGED = FALSE;
	
	// Ensure that a scan is made before proceeding, loop 1/5 sec
	while ( dig_lastqueue != CURQ ) { usleep ( 200000 ); };
	
	win_resize();
	msg_tagreset();
	win_listtitle();

}

// Displays the bottom window
void msg_show () {

	if ( !NUMMSG ) {
		wclear ( mwnd );
		mvwaddstr ( mwnd, 0, 0, "Queue is empty" );
		return;
	}
	
	mvwaddstr ( mwnd, 0, 0, "Message:" );
	mvwaddstr ( mwnd, 0, 9, queue[CURMSG].id );
	wclrtoeol ( mwnd );

	mvwaddstr ( mwnd, 1, 0, "From.:" );
	mvwaddstr ( mwnd, 1, 7, queue[CURMSG].from );
	wclrtoeol ( mwnd );

	mvwaddstr ( mwnd, 2, 0, "To...:" );
	mvwaddstr ( mwnd, 2, 7, queue[CURMSG].to );
	wclrtoeol ( mwnd );

	mvwaddstr ( mwnd, 3, 0, "Subj.:" );
	mvwaddstr ( mwnd, 3, 7, queue[CURMSG].subj );
	wclrtoeol ( mwnd );

	if ( CURQ == Q_DEFERRED ) {
		mvwaddstr ( mwnd, 4, 0, "Cause:" );
		mvwaddstr ( mwnd, 5, 0, " " );
		wclrtoeol ( mwnd );
		mvwaddstr ( mwnd, 4, 7, queue[CURMSG].cause );
		wclrtoeol ( mwnd );
	}
}

// Displays the list of messages
void queue_show() {
int i, j, f;
time_t show_start;

	j = FIRSTROW;
	i = FIRSTMSG;

	show_start = time(NULL);
	while ( j <= MAXROW && ((dig_limit && time(NULL)-show_start<=dig_limit) || !dig_limit) ) {

		if ( i<NUMMSG ) {
			f = msg_fetch ( i );
			if ( queue[i].tagged )
				wattron ( lwnd, A_BOLD );
			if ( i == CURMSG || i == MSGMARK )
				wattron ( lwnd, A_REVERSE );
			mvwaddstr  ( lwnd, j, 0, bline );
			mvwaddnstr ( lwnd, j, 1, queue[i].id, sizeof(queue[i].id) );
			mvwaddnstr ( lwnd, j, COL1,   queue[i].from, sizeof(queue[i].from) );
			mvwaddstr  ( lwnd, j, COL2-2, "  " );
			mvwaddnstr ( lwnd, j, COL2,   queue[i].to, sizeof(queue[i].to) );
			wattroff ( lwnd, A_BOLD | A_REVERSE );
		} else {
			mvwaddstr ( lwnd, j, 1, " " );
			wclrtoeol ( lwnd );
		}
		j++;
		i++;
	}

	if ( NUMMSG && CURMSG > (NUMMSG-1) ) {
		goto_last();
	}
	keypad ( lwnd, 1 );
	win_listtitle();
	show_cursor();
	msg_show();
	wnoutrefresh ( mwnd );
	wnoutrefresh ( lwnd );
	doupdate();
}

// Search for a message; returns -1 if not found
int msg_match(int s, int direction) {
int i, res;

	res = 0;

	if ( direction==0 ) {
		for ( i=s; i<NUMMSG; i++ ) {
			msg_fetch ( i );
			if ( ( (search_mode & SM_FROM) && regexec(regexp,queue[i].from,res,NULL,0)==0 ) ||
			     ( (search_mode & SM_TO)   && regexec(regexp,queue[i].to  ,res,NULL,0)==0 ) ||
			     ( (search_mode & SM_SUBJ) && regexec(regexp,queue[i].subj,res,NULL,0)==0 ) ) {
				goto_msg ( i );
				queue_show();
				return i;
			}
		}
	} else {
		for ( i=s; i>=0; i-- ) {
			msg_fetch ( i );
			if ( ( (search_mode & SM_FROM) && regexec(regexp,queue[i].from,res,NULL,0)==0 ) ||
			     ( (search_mode & SM_TO)   && regexec(regexp,queue[i].to  ,res,NULL,0)==0 ) ||
			     ( (search_mode & SM_SUBJ) && regexec(regexp,queue[i].subj,res,NULL,0)==0 ) ) {
				goto_msg ( i );
				queue_show();
				return i;
			}
		}
	}
	return -1;
}

void msg_search() {
int res;

	wnd_getregexp();
	res = regcomp ( regexp, regexps, 0 );
	if ( !res ) 
		msg_match(CURMSG,0);
}

int msg_searchnext() {
	if ( !strlen(regexps) )
		return -1;
	return msg_match(CURMSG+1,0);
}

int msg_searchprev() {
	if ( !strlen(regexps) )
		return -1;
	return msg_match(CURMSG-1,1);
}

void msg_searchandtag() {
int res;

	wnd_getregexp();
	res = regcomp ( regexp, regexps, 0 );
	if ( !res ) {
		res = msg_match(0,0);
		while ( res != -1 ) {
			queue[res].tagged = 1;
			NUMTAG++;
			res = msg_searchnext(res,0);
		}
	}
}

// Toggle envelope usage for from/to
void toggle_envelope() {
	use_envelope = !use_envelope;
	msg_cachereset();
	win_resize();
	queue_show();
}

// Main loop
void main_loop() {
int c;
int i;

	c = 0;
	
	CURMSG = 0;
	FIRSTMSG = 0;

	wnd_clear ( mwndb, "Message" );

	while ( c != 'q' ) {

		queue_show();
		c = wgetch ( lwnd );
		PREVROW = CURROW;
		
		switch ( c ) {
		case 'G':
		case KEY_END:
			goto_last();
			break;
		case 'g':
		case KEY_HOME:
			goto_first();
			break;
		case KEY_UP:
			goto_prev();
			break;
		case KEY_DOWN:
			goto_next();
			break;
		case KEY_PPAGE:
			goto_prevpage();
			break;
		case KEY_NPAGE:
			goto_nextpage();
			break;
		case KEY_RESIZE:
			win_resize();
			break;
		case 'e':
			toggle_envelope();
			break;
		case 'd':
			msg_action( MSG_DELETE );
			break;
		case 'h':
			msg_action( MSG_HOLD );
			break;
		case 'l':
			if ( CURQ == Q_HOLD )
				msg_action ( MSG_RELEASE );
			break;
		case 'r':
			msg_action ( MSG_REQUEUE );
			break;
		case '1':
			queue_change ( Q_DEFERRED );
			break;
		case '4':
			queue_change ( Q_HOLD );
			break;
		case '3':
			queue_change ( Q_INCOMING );
			break;
		case '2':
			queue_change ( Q_ACTIVE );
			break;
		case 'a':
			msg_tagall();
			break;
		case '.':
			MSGMARK = CURMSG;
			queue[CURMSG].tagged = TRUE;
			queue_show();
			break;
		case 't':
			if ( MSGMARK==-1 ) {
				if ( queue[CURMSG].tagged )
					NUMTAG--;
				else
					NUMTAG++;
				queue[CURMSG].tagged = !queue[CURMSG].tagged;
			} else {
				if ( CURMSG>=MSGMARK ) {
					for ( i=MSGMARK; i<=CURMSG; i++ ) {
						queue[i].tagged = TRUE;
					}
				} else {
					for ( i=CURMSG; i<=MSGMARK; i++ ) {
						queue[i].tagged = TRUE;
					}
				}
			}
			MSGMARK = -1;
			queue_show();
			break;
		case 'T':
			msg_searchandtag();
			break;
		case 'u':
			msg_tagreset();
			queue_show();
			break;
		case ';':
			WRK_TAGGED = !WRK_TAGGED;
			queue_show();
			break;
		case '/':
			msg_search();
			break;
		case 'n':
			msg_searchnext();
			break;
		case 'p':
			msg_searchprev();
			break;
		case '?':
			help();
			break;
		case 'c':
			ASK_CONFIRM = !ASK_CONFIRM;
			break;
		case '-':
			DO_SCAN = !DO_SCAN;
			break;
		case 'q':
		case 'Q':
			break;
		case 's':
		case 'S':
			if ( NUMMSG ) {
				msg_cat();
				win_resize();
				msg_show();
			}
			break;
		}
		
	}
}

int main ( int argc, char** argv ) {
	FILE *p;
	char pconf[BUF_SIZE];
	char buf[BUF_SIZE];
	int opt;
	int hd, initial_queue;

	// Some defaults
	MSGMARK = -1;
	ASK_CONFIRM = TRUE;
	MSG_MAX = 200;
	FIRSTROW = 2;
	hd = 2;
	use_envelope = FALSE;
	ISROOT = (getuid()==0);
	pf_version = -1;
	initial_queue = 0;
	dig_suspend = 0;
	dig_limit = 0;
	DO_SCAN = 1;

	// Init mutex
	pthread_mutex_init ( &dig_mutex, NULL );

	memset ( config_path, 0, sizeof(config_path) );
	memset ( pftools_path, 0, sizeof(pftools_path) );
	memset ( postconf_path, 0, sizeof(postconf_path) );
	memset ( postsuper_path, 0, sizeof(postsuper_path) );
	memset ( postcat_path, 0, sizeof(postcat_path) );

	regexps = (char*) malloc ( REGEXP_LEN );
	regexp  = (regex_t*) malloc (sizeof(regex_t));

	// Look for other options
	opt = 0;
	while ( opt!=-1 ) {
		opt = getopt ( argc, argv, "q:p:c:s:m:l:ehvr" );
		switch (opt) {
			case 'q':
				initial_queue = atoi ( optarg );
				switch ( initial_queue ) {
				case 1:
					initial_queue = 0;
					break;
				case 2:
					initial_queue = 3;
					break;
				case 3:
					initial_queue = 2;
					break;
				case 4:
					initial_queue = 1;
					break;
				default:
					initial_queue = 0;
				}
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'v':
				version();
				exit(0);
				break;
			case 'm':
				MSG_MAX = atoi ( optarg );
				if ( MSG_MAX > 2048 )
					MSG_MAX = 2048;
				if ( MSG_MAX < 5 )
					MSG_MAX = 5;
				break;
			case 's':
				hd = atoi ( optarg );
				if ( hd > 300 )
					hd = 300;
				if ( hd < 1 )
					hd = 1;
				break;
			case 'l':
				dig_limit = atoi ( optarg );
				if ( dig_limit > 300 )
					dig_limit = 300;
				if ( dig_limit < 1 )
					dig_limit = 1;
				break;
			case 'e':
				use_envelope = TRUE;
				break;
			case 'p':
				if ( strlen(optarg) ) {
					snprintf ( pftools_path, BUF_SIZE-1, "%s", optarg );
				}
				break;
			case 'c':
				if ( strlen(optarg) ) {
					snprintf ( config_path, BUF_SIZE-1, "%s", optarg );
				}
				break;
			case 'r':
				if ( strlen(optarg) ) {
					switch ( optarg[0] ) {
					case 'x':
						pf_version = PF_VERSION_1x;
						break;
					case '0':
						pf_version = PF_VERSION_20;
						break;
					case '1':
						pf_version = PF_VERSION_21;
						break;
					case '2':
						pf_version = PF_VERSION_22;
						break;
					}
				}
				break;
		}
	}

	// If -p is not specified, use commands without path
	if ( strlen ( pftools_path ) ) {
		snprintf ( postconf_path, BUF_SIZE, "%s/postconf", pftools_path );
		snprintf ( postsuper_path, BUF_SIZE, "%s/postsuper", pftools_path );
		snprintf ( postcat_path, BUF_SIZE, "%s/postcat", pftools_path );
	} else {
		snprintf ( postconf_path, BUF_SIZE, "postconf" );
		snprintf ( postsuper_path, BUF_SIZE, "postsuper" );
		snprintf ( postcat_path, BUF_SIZE, "postcat" );
	}
	
	if ( pf_version == -1 ) {
		// Version autodetect
		pf_version = PF_VERSION_22;

		has_configpath = strlen(config_path);
	
		if ( has_configpath )
			snprintf ( pconf, BUF_SIZE, "%s -c %s -h mail_version 2> /dev/null", postconf_path, config_path );
		else
			snprintf ( pconf, BUF_SIZE, "%s -h mail_version 2> /dev/null", postconf_path );
		p = popen ( pconf, "r" );
		if ( !p ) {
			fprintf ( stderr, "Cannot guess postfix version, using 2.2 as default\n" );
			sprintf ( buf, "2.2" );
		} else
		freadl ( p, buf, sizeof(buf) );
		if ( !strncmp ( buf, "1", 1 ) )
			pf_version = PF_VERSION_1x;
		if ( !strncmp ( buf, "2.0", 3 ) )
			pf_version = PF_VERSION_20;
		if ( !strncmp ( buf, "2.1", 3 ) )
			pf_version = PF_VERSION_21;
		if ( !strncmp ( buf, "2.2", 3 ) )
			pf_version = PF_VERSION_22;
	}
	
	// Look for queue_directory
	if ( has_configpath ) 
		snprintf ( pconf, BUF_SIZE, "%s -c %s -h queue_directory 2> /dev/null", postconf_path, config_path );
	else
		snprintf ( pconf, BUF_SIZE, "%s -h queue_directory 2> /dev/null", postconf_path );
	p = popen ( pconf, "r" );
	if ( !p ) {
		fprintf ( stderr, "Cannot use postconf to search queue_directory, command was:\n%s\n", pconf );
		exit (-2);
	}
	if ( !freadl ( p, queue_path, sizeof(queue_path) ) ) {
		fprintf ( stderr, "Cannot use postconf to search queue_directory, command was:\n%s\n", pconf );
		exit (-1);
	}
	pclose ( p );

	// Look for has_queue_depth
	if ( has_configpath )
		snprintf ( buf, BUF_SIZE, "%s -c %s -h hash_queue_depth 2> /dev/null", postconf_path, config_path );
	else
		snprintf ( buf, BUF_SIZE, "%s -h hash_queue_depth 2> /dev/null", postconf_path );
	p = popen ( buf, "r" );
	if ( !p ) {
		fprintf ( stderr, "Cannot use postconf to search hash_queue_depth, using 1 as default\n" );
		queue_depth = 1;
	} else {
		freadl ( p, buf, sizeof(buf) );
		queue_depth = atoi(buf);
	}
	
	NUMTAG = 0;
	WRK_TAGGED = FALSE;
	FIRSTMSG = 0;
	mainwnd = NULL;
	mwndb = lwnd = mwnd = NULL;
	bline = NULL;

	queue = (struct msg_t*)malloc(sizeof(struct msg_t)*MSG_MAX);
	if ( !queue ) {
		fprintf ( stderr, "Sorry, cannot malloc for %d elements!\n", MSG_MAX );
		exit (-2);
	}
	queue_thread = (struct msg_t*)malloc(sizeof(struct msg_t)*MSG_MAX);
	if ( !queue_thread ) {
		free ( queue );
		fprintf ( stderr, "Sorry, cannot malloc for %d elements!\n", MSG_MAX );
		exit (-2);
	}

	win_resize();

	halfdelay ( hd*10 );
	delay_time = hd;
	noecho();
	
	dig_lastqueue = -1;
	fill_list();
	queue_change ( initial_queue );

	fill_list();
	main_loop();

	pthread_mutex_destroy ( &dig_mutex );

	free ( queue );
	free ( queue_thread );
	free ( bline );
	free ( regexps );
	regfree ( regexp );

	endwin();

	return 0;

}
